"use strict";

exports.__esModule = true;
exports.default = babelPresetModules;

var _path = _interopRequireDefault(require("path"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * A preset that transforms unweildy transpiled code into modern source that engines and bundlers can better optimize.
 * @param {Object} [options]
 * @param {boolean} [options.loose=false] Loose mode enables transforms that produce smaller output but may cause issues with certain code.
 * @param {boolean} [options.module=true] Assume code will be run in a Strict Mode or ES Modules environment.
 * @param {boolean} [options.webpack=true] Parse and explode Webpack bundles
 */
function babelPresetModules(_, options = {}) {
  const loose = options.loose === true;
  const module = options.module !== false;
  const webpack = options.webpack !== false;
  return {
    plugins: [_path.default.resolve(__dirname, './plugins/transform-umd-to-cjs'), // Currently disabled, causes problems with non-CJS files:
    // path.resolve(__dirname, './plugins/transform-cjs-to-esm'),
    [_path.default.resolve(__dirname, './plugins/transform-arguments'), {
      loose
    }], _path.default.resolve(__dirname, './plugins/transform-destructuring'), _path.default.resolve(__dirname, './plugins/transform-classes'), _path.default.resolve(__dirname, './plugins/transform-methods'), // too broken:
    // path.resolve(__dirname, './plugins/transform-mangle-identifiers'),
    _path.default.resolve(__dirname, './plugins/transform-fallbacks'), _path.default.resolve(__dirname, './plugins/transform-iife-arrows'), _path.default.resolve(__dirname, './plugins/transform-implicit-scope'), _path.default.resolve(__dirname, './plugins/transform-array-spread'), _path.default.resolve(__dirname, './plugins/transform-template-literals'), webpack && _path.default.resolve(__dirname, './plugins/transform-optimize-webpack'), _path.default.resolve(__dirname, './plugins/transform-remove-polyfills'), module && _path.default.resolve(__dirname, './plugins/transform-implicit-strict')].filter(Boolean)
  };
}

module.exports = exports.default;